import os
import json
import shutil
import traceback
import uuid
from typing import Callable, Any

from petdb import PetDB, PetCollection, NON_EXISTENT, NonExistent

TEST_COLLECTION_NAME = "testcol"

RED = "\033[91m"
GREEN = "\033[92m"
COLOREND = "\033[0m"

with open(__file__, "r") as f:
	testfile = f.read().splitlines()

def get_line(func: Callable):
	search = f"def {func.__name__}(col: PetCollection):"
	for i, line in enumerate(testfile, 1):
		if search in line:
			return i + 1
	return 0

class Tests:

	root: str
	db: PetDB
	col: PetCollection
	sets: list[Callable] = []
	tests: list = []

	@classmethod
	def init(cls):
		cls.root = os.path.join(os.getcwd(), "testdb")
		if os.path.exists(cls.root):
			shutil.rmtree(cls.root)
		os.mkdir(cls.root)
		cls.db = PetDB(cls.root)

	@classmethod
	def setup(cls, init: list, ignore_id: bool):
		cls.db.delete(TEST_COLLECTION_NAME)
		if not ignore_id:
			for doc in init:
				if "_id" not in doc:
					doc["_id"] = str(uuid.uuid4())
		with open(os.path.join(cls.root, "petstorage", f"{TEST_COLLECTION_NAME}.json"), "w") as f:
			json.dump(init, f)
		cls.col = cls.db.create_collection(TEST_COLLECTION_NAME)

	@classmethod
	def run(cls, nums: int | list[int] = None):
		if isinstance(nums, int):
			nums = [nums]
		for testset in cls.sets:
			testset()
		passed = 0
		failed = []
		for i, test in enumerate(cls.tests, 1):
			try:
				if nums is not None and i not in nums:
					continue
				cls.setup(test["init"], test["ignore_id"])
				result = test["func"](cls.col)
				if not test["ignore_id"]:
					cls.remove_ids(result)
					cls.remove_ids(test["expect"])
				if result == test["expect"]:
					passed += 1
				else:
					failed.append(cls.format_failed(i, test, result))
			except Exception as e:
				if isinstance(test["expect"], Exception) and test["expect"].__class__ == e.__class__:
					passed += 1
				else:
					failed.append(cls.format_exception(i, test))
		shutil.rmtree(cls.root)
		return passed, failed

	@classmethod
	def format_failed(cls, number: int, test: dict, result: Any) -> str:
		return (f"{RED}Test {number} ({test["name"]}){COLOREND}: expected '{test["expect"]}', got '{result}'\n"
			f"File \"{__file__}\", line {get_line(test["func"])}")

	@classmethod
	def format_exception(cls, number: int, test: dict) -> str:
		return (f"{RED}Exception raised during Test {number} ({test["name"]}):{COLOREND}\n"
			+ traceback.format_exc().replace("\n\n", "\n"))

	@classmethod
	def remove_ids(cls, entry):
		if isinstance(entry, dict):
			entry.pop("_id", None)
		elif isinstance(entry, list) and len(entry) > 0 and isinstance(entry[0], dict):
			for doc in entry:
				doc.pop("_id", None)

def test(init=None, expect=None, ignore_id: bool = False, init_expectation: bool = True):
	if init is None and expect and init_expectation:
		if isinstance(expect, dict):
			init = [expect]
		elif isinstance(expect, list) and all(isinstance(doc, dict) for doc in expect):
			init = expect[:]
	def decorator(func):
		Tests.tests.append({
			"func": func,
			"init": init or [],
			"expect": expect,
			"ignore_id": ignore_id,
			"name": func.__name__})
	return decorator

def testset(func):
	Tests.sets.append(func)

# BASE

@testset
def base():

	@test(init=[{"a": 1}, {"a": 2}, {"a": 3}], expect=[{"a": 1}, {"a": 2}, {"a": 3}])
	def collection_to_list(col: PetCollection):
		return list(col)

	@test(init=[{"a": 1}, {"a": 2}, {"a": 3}], expect=[{"a": 1}, {"a": 2}, {"a": 3}])
	def iteration1(col: PetCollection):
		result = []
		for doc in col:
			result.append(doc)
		return result

	@test(init=[{"a": 1}, {"a": 2}, {"a": 3}], expect=[1, 2, 3])
	def iteration2(col: PetCollection):
		result = []
		for doc in col:
			result.append(doc["a"])
		return result

# INSERT

@testset
def insertion():

	@test(expect=[{"a": 5}], init_expectation=False)
	def insert_one(col: PetCollection):
		col.insert({"a": 5})
		return list(col)

	@test(expect=[{"a": 1}, {"a": 2}, {"a": 3}], init_expectation=False)
	def insert_many(col: PetCollection):
		col.insert({"a": 1})
		col.insert({"a": 2})
		col.insert({"a": 3})
		return list(col)

	@test(expect=True)
	def id_generation(col: PetCollection):
		doc1 = col.insert({"a": 1})
		doc2 = col.insert({"a": 2})
		return ("_id" in doc1 and len(doc1["_id"]) > 8
				and "_id" in doc2 and len(doc2["_id"]) > 8
				and doc1["_id"] != doc2["_id"])

	@test(expect=[{"a": 5}], init_expectation=False)
	def dump_on_insert(col: PetCollection):
		col.insert({"a": 5})
		with open(os.path.join(Tests.root, "petstorage", f"{TEST_COLLECTION_NAME}.json")) as f:
			return json.load(f)

# SELECTION

@testset
def selection():

	@test(expect={"a": 1}, init_expectation=False)
	def get_by_id1(col: PetCollection):
		doc_id = col.insert({"a": 1})["_id"]
		return col.get(doc_id)

	@test(expect=None)
	def get_by_id2(col: PetCollection):
		return col.get("12345")

	@test(expect={"a": 5, "b": 10})
	def find1(col: PetCollection):
		return col.find({"a": 5})

	@test(init=[{"a": 5, "b": 10}], expect=None)
	def find2(col: PetCollection):
		return col.find({"a": 10})

	@test(expect={"a": 5, "b": 10, "c": 20})
	def find3(col: PetCollection):
		return col.find({"a": 5, "c": 20})

	@test(expect={"a": 5, "b": 10, "c": {"d": 20}})
	def find4(col: PetCollection):
		return col.find({"a": 5, "c.d": 20})

	@test(expect={"a": 5, "b": 10, "c": {"d": 20}})
	def find5(col: PetCollection):
		return col.find({"a": 5, "c": {"d": 20}})

	@test(expect={"a": 5, "b": 10, "c": {"d": {"e": 20}}})
	def find6(col: PetCollection):
		return col.find({"a": 5, "c": {"d.e": 20}})

	@test(expect={"a": 5, "b": 10, "c": {"d": {"e": 20}}})
	def find6(col: PetCollection):
		return col.find({"a": 5, "c.d": {"e": 20}})

	@test(expect={"a": 5, "b": 10, "c": {"d": {"e": 20}}})
	def find6(col: PetCollection):
		return col.find({"a": 5, "c": {"d": {"e": 20}}})

	@test(expect=[{"a": 5, "b": 1, "c": 3}, {"a": 6, "b": 1}, {"a": 7, "b": 1}])
	def findall1(col: PetCollection):
		return col.findall({})

	@test(expect=[{"a": 5, "b": 1, "c": 3}, {"a": 6, "b": 1}, {"a": 7, "b": 1}])
	def findall2(col: PetCollection):
		return list(col.findall({}))

	@test(init=[{"a": 5, "c": 3}, {"a": 6, "b": 1}, {"a": 7, "b": 1}], expect=[{"a": 6, "b": 1}, {"a": 7, "b": 1}])
	def findall3(col: PetCollection):
		return col.findall({"b": 1})

	@test(init=[{"a": 5, "b": 2, "c": 3}, {"a": 6, "b": 1}, {"a": 7, "b": 1}], expect=[{"a": 6, "b": 1}, {"a": 7, "b": 1}])
	def findall4(col: PetCollection):
		return col.findall({"b": 1})

	@test(init=[{"a": 5, "b": 2, "c": 3}, {"a": 6, "b": 1}, {"a": 7, "b": 1}], expect=[])
	def findall5(col: PetCollection):
		return col.findall({"b": 3})

	@test(init=[{"a": [{}, {"b": 1}]}, {"a": [{}, {"b": 2}]}, {"a": [{}, {"b": 3}]}, {"a": []}],
		expect=[{"a": [{}, {"b": 1}]}, {"a": [{}, {"b": 2}]}])
	def findall6(col: PetCollection):
		return col.filter({"a.1.b": {"$exists": True}}).findall({"a.1.b": {"$lt": 3}})

# OPERATORS

@testset
def operators():

	@test(expect={"a": 5, "b": 10})
	def eq1(col: PetCollection):
		return col.find({"a": {"$eq": 5}})

	@test(init=[{"a": 5, "b": 10}], expect=None)
	def eq2(col: PetCollection):
		return col.find({"a": {"$eq": 50}})

	@test(expect={"a": 5, "b": 10})
	def not1(col: PetCollection):
		return col.find({"a": {"$not": 1}})

	@test(init=[{"a": 5, "b": 10}], expect=None)
	def not2(col: PetCollection):
		return col.find({"a": {"$not": 5}})

	@test(expect={"a": 5, "b": 10})
	def ne1(col: PetCollection):
		return col.find({"a": {"$ne": 1}})

	@test(init=[{"a": 5, "b": 10}], expect=None)
	def ne2(col: PetCollection):
		return col.find({"a": {"$ne": 5}})

	@test(init=[{"a": 5, "b": 10}, {"a": 10, "b": 5}, {"a": 1, "b": 0}], expect=[{"a": 1, "b": 0}])
	def lt(col: PetCollection):
		return col.findall({"a": {"$lt": 5}})

	@test(init=[{"a": 5, "b": 10}, {"a": 10, "b": 5}, {"a": 1, "b": 0}], expect=[{"a": 5, "b": 10}, {"a": 1, "b": 0}])
	def lte(col: PetCollection):
		return col.findall({"a": {"$lte": 5}})

	@test(init=[{"a": 5, "b": 10}, {"a": 10, "b": 5}, {"a": 1, "b": 0}], expect=[{"a": 10, "b": 5}])
	def gt(col: PetCollection):
		return col.findall({"a": {"$gt": 5}})

	@test(init=[{"a": 5, "b": 10}, {"a": 10, "b": 5}, {"a": 1, "b": 0}], expect=[{"a": 5, "b": 10}, {"a": 10, "b": 5}])
	def gte(col: PetCollection):
		return col.findall({"a": {"$gte": 5}})

	@test(init=[{"a": 5, "b": 10}, {"a": 10, "b": 5}, {"a": 1, "b": 0}], expect=[{"a": 5, "b": 10}])
	def gt_an_lt(col: PetCollection):
		return col.findall({"a": {"$gt": 3, "$lt": 7}})

	@test(init=[{"a": 5, "b": 10}, {"a": 10, "b": 5}, {"a": 1, "b": 0}], expect=[{"a": 5, "b": 10}])
	def in1(col: PetCollection):
		return col.findall({"a": {"$in": [4, 5, 6]}})

	@test(init=[{"a": 5, "b": 10}, {"a": 10, "b": 5}, {"a": 1, "b": 0}], expect=[])
	def in2(col: PetCollection):
		return col.findall({"a": {"$in": [4, 6]}})

	@test(init=[{"a": 5, "b": 10}, {"a": 10, "b": 5}, {"a": 1, "b": 0}], expect=[{"a": 10, "b": 5}])
	def nin1(col: PetCollection):
		return col.findall({"a": {"$nin": [1, 4, 5, 6]}})

	@test(init=[{"a": 5, "b": 10}, {"a": 10, "b": 5}, {"a": 1, "b": 0}], expect=[])
	def nin2(col: PetCollection):
		return col.findall({"a": {"$nin": [1, 4, 5, 6, 10]}})

	@test(init=[{"a": 5, "b": 10}, {"a": 10, "b": 5}, {"a": 1, "b": 0}], expect=[{"a": 10, "b": 5}])
	def notin1(col: PetCollection):
		return col.findall({"a": {"$notin": [1, 4, 5, 6]}})

	@test(init=[{"a": 5, "b": 10}, {"a": 10, "b": 5}, {"a": 1, "b": 0}], expect=[])
	def notin2(col: PetCollection):
		return col.findall({"a": {"$notin": [1, 4, 5, 6, 10]}})

	@test(expect=[{"a": 5}])
	def exists1(col: PetCollection):
		return col.findall({"a": {"$exists": True}})

	@test(init=[{"a": 5}], expect=[])
	def exists2(col: PetCollection):
		return col.findall({"b": {"$exists": True}})

	@test(init=[{"a": 5}, {"a": 6}, {"b": 10}], expect=[{"a": 5}, {"a": 6}])
	def exists3(col: PetCollection):
		return col.findall({"a": {"$exists": True}})

	@test(init=[{"a": 5}, {"a": 6}, {"b": 10}], expect=[{"b": 10}])
	def exists4(col: PetCollection):
		return col.findall({"a": {"$exists": False}})

	@test(expect=[{"a": 5}, {"a": 6}, {"b": 10}])
	def exists5(col: PetCollection):
		return col.findall({"a.b": {"$exists": False}})

	@test(init=[{"a": 5}, {"a": 6}, {"b": 10}], expect=[])
	def exists6(col: PetCollection):
		return col.findall({"a.b": {"$exists": True}})

	@test(expect=[{"a": "12345"}, {"a": "67890"}])
	def regex1(col: PetCollection):
		return col.findall({"a": {"$regex": "\\d{5}"}})

	@test(init=[{"a": "12345"}, {"a": "67890"}], expect=[])
	def regex2(col: PetCollection):
		return col.findall({"a": {"$regex": "\\D{5}"}})

	@test(init=[{"a": "12345"}, {"a": "678901"}], expect=[{"a": "12345"}])
	def regex3(col: PetCollection):
		return col.findall({"a": {"$regex": "^1"}})

	@test(init=[{"a": "12345"}, {"a": "678901"}], expect=[{"a": "12345"}])
	def regex4(col: PetCollection):
		return col.findall({"a": {"$regex": "5$"}})

	@test(init=[{"a": "12345"}, {"a": "abcde"}], expect=[{"a": "12345"}])
	def func1(col: PetCollection):
		return col.findall({"a": {"$func": str.isdigit}})

	@test(init=[{"a": "12345"}, {"a": "abcde"}, {"a": 12345}], expect=TypeError())
	def func2(col: PetCollection):
		return col.findall({"a": {"$func": str.isdigit}})

	@test(init=[{"a": "12345"}, {"a": "abcde"}], expect=[{"a": "12345"}])
	def func3(col: PetCollection):
		return col.findall({"a": {"$func": str.isdigit}})

	@test(init=[{"a": "12345"}, {"a": "abcde"}, {"a": 12345}], expect=[{"a": "12345"}, {"a": "abcde"}])
	def type1(col: PetCollection):
		return col.findall({"a": {"$type": str}})

	@test(expect=[{"a": "12345"}, {"a": "abcde"}, {"a": 12345}])
	def type2(col: PetCollection):
		return col.findall({"b": {"$type": NonExistent}})

# LOGICAL OPERATORS

@testset
def logical_operators():

	@test(init=[{"a": 1}, {"a": 2}, {"a": 3}, {"a": 4}], expect=[{"a": 1}, {"a": 3}])
	def or1(col: PetCollection):
		return col.findall({"$or": [{"a": 1}, {"a": 3}]})

	@test(init=[{"a": 1}, {"a": 2}, {"a": 3}, {"a": 4}], expect=[{"a": 1}, {"a": 3}])
	def or2(col: PetCollection):
		return col.findall({"a": {"$or": [1, 3]}})

	@test(init=[{"a": 1}, {"a": 2}, {"a": 3}, {"a": 4}], expect=[{"a": 1}, {"a": 3}])
	def or3(col: PetCollection):
		return col.findall({"a": {"$or": [{"$lt": 2}, {"$eq": 3}]}})

	@test(init=[{"a": {"b": 1}}, {"a": {"b": 7}}, {"a": {"b": 10}}], expect=[{"a": {"b": 1}}, {"a": {"b": 10}}])
	def or4(col: PetCollection):
		return col.findall({"$or": [{"a.b": 1}, {"a.b": 10}]})

	@test(init=[{"a": {"b": 1}}, {"a": {"b": 7}}, {"a": {"b": 10}}], expect=[{"a": {"b": 1}}, {"a": {"b": 10}}])
	def or5(col: PetCollection):
		return col.findall({"$or": [{"a.b": {"$lt": 5}}, {"a.b": {"$gte": 10}}]})

	@test(init=[{"a": 1}, {"a": 2}, {"a": 3}, {"a": 4}], expect=[{"a": 2}])
	def and1(col: PetCollection):
		return col.findall({"$and": [{"a": {"$gt": 1}}, {"a": {"$lt": 3}}]})

	@test(init=[{"a": 1}, {"a": 2}, {"a": 3}, {"a": 4}, {"b": 2}], expect=[{"a": 2}])
	def and2(col: PetCollection):
		return col.findall({"a": {"$and": [{"$exists": True}, {"$gt": 1}, {"$lt": 3}]}})

# DELETION

@testset
def deletion():

	@test(init=[{"a": 5}, {"b": 2}, {"b": 3}], expect=[])
	def remove_clear1(col: PetCollection):
		col.remove({})
		return col.list()

	@test(expect=[{"a": 5}, {"b": 2}, {"b": 3}])
	def remove_clear2(col: PetCollection):
		return col.remove({})

	@test(expect=[], init=[{"a": 5}, {"b": 2}, {"b": 3}])
	def remove_clear3(col: PetCollection):
		col.remove()
		return col.list()

	@test(expect=[{"a": 5}, {"b": 2}, {"b": 3}])
	def remove_clear4(col: PetCollection):
		return col.remove()

	@test(expect=[{"a": 5}, {"b": 2}], init=[{"a": 5}, {"b": 2}, {"b": 3}])
	def remove1(col: PetCollection):
		col.remove({"b": 3})
		return col.list()

	@test(expect=[{"b": 2}], init=[{"a": 5}, {"b": 2}, {"b": 3}])
	def remove2(col: PetCollection):
		return col.remove({"b": 2})

	@test(expect=[{"a": 5}, {"b": 3}], init=[{"a": 5}, {"b": 2}, {"b": 3}])
	def remove3(col: PetCollection):
		col.remove(col.flat("_id")[1])
		return col.list()

	@test(expect=[{"b": 2}], init=[{"a": 5}, {"b": 2}, {"b": 3}])
	def remove4(col: PetCollection):
		return col.remove(col.flat("_id")[1])

	@test(expect=[{"a": 5}, {"b": 3}], init=[{"a": 5}, {"b": 2}, {"b": 3}])
	def remove5(col: PetCollection):
		col.remove([col.flat("_id")[1]])
		return col.list()

	@test(expect=[{"b": 2}], init=[{"a": 5}, {"b": 2}, {"b": 3}])
	def remove6(col: PetCollection):
		return col.remove([col.flat("_id")[1]])

	@test(expect=[{"a": 5}], init=[{"a": 5}, {"b": 2}, {"b": 3}])
	def remove7(col: PetCollection):
		col.remove(col[1:])
		return col.list()

	@test(expect=[{"b": 2}, {"b": 3}], init=[{"a": 5}, {"b": 2}, {"b": 3}])
	def remove8(col: PetCollection):
		return col.remove(col[1:])

# SORTING

@testset
def sorting():

	@test(init=[{"a": 5}, {"a": 4}, {"a": 8}, {"a": 10}, {"a": 2}, {"a": 1}],
		expect=[{"a": 1}, {"a": 2}, {"a": 4}, {"a": 5}, {"a": 8}, {"a": 10}])
	def sort1(col: PetCollection):
		return col.sort("a").list()

	@test(init=[{"a": 5}, {"a": 4}, {"a": 8}, {"b": 0}, {"a": 10}, {"a": 2}, {"a": 1}],
		expect=[{"a": 1}, {"a": 2}, {"a": 4}, {"a": 5}, {"a": 8}, {"a": 10}, {"b": 0}])
	def sort2(col: PetCollection):
		return col.sort("a").list()

	@test(init=[{"a": "5"}, {"a": 4}, {"a": "8"}, {"b": 0}, {"a": 10}, {"a": 2}, {"a": 1}],
		expect=[{'a': 1}, {'a': 2}, {'a': 4}, {'a': 10}])
	def sort3(col: PetCollection):
		return col.filter({"a": {"$type": int}}).sort("a").list()

	@test(init=[{"a": "5"}, {"a": 4}, {"a": "8"}, {"b": 0}, {"a": 10}, {"a": 2}, {"a": 1}],
		expect=[{'a': 10}, {'a': 4}, {'a': 2}, {'a': 1}])
	def sort4(col: PetCollection):
		return col.filter({"a": {"$type": int}}).sort("a", reverse=True).list()

	@test(init=[{"a": 5}, {"a": 4}, {"a": 8}, {"a": 10}, {"a": 2}, {"a": 1}],
		expect=[[1], [2], [4], [5], [8], [10]])
	def sort5(col: PetCollection):
		return col.map(lambda doc: [doc["a"]]).sort(0).list()

	@test(init=[{"a": 5}, {"a": 4}, {"a": 8}, {"a": 10}, {"a": 2}, {"a": 1}],
		expect=[[1], [2], [4], [5], [8], [10]])
	def sort6(col: PetCollection):
		return col.map(lambda doc: [doc["a"]]).sort("0").list()

	@test(init=[{"a": 5}, {"a": 4}, {"a": 8}, {"b": 0}, {"a": 10}, {"a": 2}, {"a": 1}],
		expect=[[1], [2], [4], [5], [8], [10], [None]])
	def sort7(col: PetCollection):
		return col.map(lambda doc: [doc.get("a", None)]).sort(lambda i: (i[0] is None, i[0])).list()

	@test(init=[{"a": 5}, {"a": 4}, {"a": 8}, {"a": 10}, {"a": 2}, {"a": 1}],
		expect=[[[0, 1]], [[0, 2]], [[0, 4]], [[0, 5]], [[0, 8]], [[0, 10]]])
	def sort8(col: PetCollection):
		return col.map(lambda doc: [[0, doc.get("a", None)]]).sort("0.1").list()

	@test(init=[{"a": 5}, {"a": 4}, {"a": 8}, {"b": 0}, {"a": 10}, {"a": 2}, {"a": 1}],
		expect=[1, 2, 4, 5, 8, 10, NON_EXISTENT])
	def sort9(col: PetCollection):
		return col.map(lambda doc: doc.get("a", NON_EXISTENT)).sort().list()

	@test(init=[{"a": 5, "b": 5}, {"a": 5, "b": 3}, {"a": 4, "b": 4}],
		expect=[{"a": 4, "b": 4}, {"a": 5, "b": 3}, {"a": 5, "b": 5}])
	def sort10(col: PetCollection):
		return col.sort(("a", "b")).list()

# MAPPING AND MUTABLE MAPPING

@testset
def mapping():

	@test(init=[{"a": 5}, {"a": 10}], expect=[5, 10])
	def map1(col: PetCollection):
		return col.map(lambda doc: doc["a"]).list()

	@test(init=[{"a": 5}, {"b": 1}, {"a": 10}], expect=[25, 0, 100])
	def map2(col: PetCollection):
		return col.map(lambda doc: doc.get("a", 0) ** 2).list()

	@test(init=[{"a": 5}, {"b": 1}, {"a": 10}], expect=[100, 25, 0])
	def map3(col: PetCollection):
		return col.map(lambda doc: doc.get("a", 0) ** 2).sort(reverse=True).list()

	@test(expect=[5, 6, 7, 8, 9])
	def map4(col: PetCollection):
		for i in range(10):
			col.insert({"a": i})
		return col.filter({"a": {"$gte": 5}}).map(lambda doc: doc["a"]).list()

# SIZE AND MUTABLE SIZE

@testset
def size():

	@test(expect=100)
	def size1(col: PetCollection):
		for i in range(100):
			col.insert({"a": i})
		return col.size()

	@test(expect=50)
	def size2(col: PetCollection):
		for i in range(100):
			col.insert({"a": i})
		return col.filter({"a": {"$lt": 50}}).size()

	@test(expect=0)
	def size3(col: PetCollection):
		return col.size()

	@test(expect=0)
	def size4(col: PetCollection):
		return col.filter({"a": {"$lt": 50}}).size()

# FLATTING

@testset
def flatting():

	@test(init=[{"a": 5}, {"a": 6}, {"b": 1}, {"a": 2}], expect=[5, 6, 2])
	def flat1(col: PetCollection):
		return col.flat("a").list()

	@test(init=[{"a": {"b": 5}}, {"a": {"b": 6}}, {"a": {"b": 2}}], expect=[5, 6, 2])
	def flat2(col: PetCollection):
		return col.flat("a.b").list()

	@test(init=[{"a": {"b": 5}}, {"a": {"b": 6}}, {"a": {"b": 2}}], expect=[5, 6, 2])
	def flat3(col: PetCollection):
		return col.flat("a").flat("b").list()

	@test(init=[{"a": [0, {"b": 5}]}, {"a": [0, {"b": 6}]}, {"a": [0, {"b": 2}]}], expect=[5, 6, 2])
	def flat4(col: PetCollection):
		return col.flat("a.1.b").list()

	@test(init=[{"a": [0, {"b": 5}]}, {"a": [0, {"b": 6}]}, {"a": [0, {"b": 2}]}], expect=[5, 6, 2])
	def flat5(col: PetCollection):
		return col.flat("a").flat("1.b").list()

	@test(init=[{"a": [0, {"b": 5}]}, {"a": [0, {"b": 6}]}, {"a": [0, {"b": 2}]}], expect=[5, 6, 2])
	def flat6(col: PetCollection):
		return col.flat("a.1").flat("b").list()

	@test(init=[{"a": [0, {"b": 5}]}, {"a": [0, {"b": 6}]}, {"a": [0, {"b": 2}]}], expect=[5, 6, 2])
	def flat7(col: PetCollection):
		return col.flat("a").flat(1).flat("b").list()

	@test(init=[{"a": [0, {"b": 5}]}, {"a": [0, {"b": 6}]}, {"a": [0, {"b": 2}]}], expect=[5, 6, 2])
	def flat8(col: PetCollection):
		return col.flat("a").flat("1").flat("b").list()

# MUTABLE BASE

@testset
def mutable_base():

	@test(init=[{"a": 1}, {"a": 2}, {"a": 3}], expect=[{"a": 1}, {"a": 2}, {"a": 3}])
	def collection_to_list(col: PetCollection):
		return list(col.filter({}))

	@test(init=[{"a": 1}, {"a": 2}, {"a": 3}], expect=[{"a": 1}, {"a": 2}, {"a": 3}])
	def iteration1(col: PetCollection):
		result = []
		for doc in col.filter({}):
			result.append(doc)
		return result

	@test(init=[{"a": 1}, {"a": 2}, {"a": 3}], expect=[1, 2, 3])
	def iteration2(col: PetCollection):
		result = []
		for doc in col.filter({}):
			result.append(doc["a"])
		return result

# MUTABLE DELETION

@testset
def mutable_deletion():

	@test(init=[{"a": 5, "c": 10}, {"a": 5, "c": 1}, {"a": 1, "c": 10}, {"a": 55, "c": 1}],
		expect=[{"a": 5, "c": 10}, {"a": 5, "c": 1}])
	def mutable_remove1(col: PetCollection):
		return col.filter({"a": 5}).remove()

	@test(init=[{"a": 5, "c": 10}, {"a": 5, "c": 1}, {"a": 1, "c": 10}, {"a": 55, "c": 1}],
		expect=[{"a": 1, "c": 10}, {"a": 55, "c": 1}])
	def mutable_remove2(col: PetCollection):
		col.filter({"a": 5}).remove()
		return col.list()

# MUTABLE SELECTION

@testset
def mutable_selection():

	@test(init=[{"a": 2}, {"a": 3}], expect={"a": 1})
	def mutable_get_by_id1(col: PetCollection):
		doc_id = col.insert({"a": 1})["_id"]
		return col.filter({}).get(doc_id)

	@test(expect=None)
	def mutable_get_by_id2(col: PetCollection):
		return col.filter({}).get("12345")

	@test(expect={"a": 5, "b": 10})
	def mutable_find1(col: PetCollection):
		return col.filter({}).find({"a": 5})

	@test(init=[{"a": 5, "b": 10}], expect=None)
	def mutable_find2(col: PetCollection):
		return col.filter({}).find({"a": 10})

	@test(expect={"a": 5, "b": 10, "c": 20})
	def mutable_find3(col: PetCollection):
		return col.filter({}).find({"a": 5, "c": 20})

	@test(expect={"a": 5, "b": 10, "c": {"d": 20}})
	def mutable_find4(col: PetCollection):
		return col.filter({}).find({"a": 5, "c.d": 20})

	@test(expect={"a": 5, "b": 10, "c": {"d": 20}})
	def mutable_find5(col: PetCollection):
		return col.filter({}).find({"a": 5, "c": {"d": 20}})

	@test(expect={"a": 5, "b": 10, "c": {"d": {"e": 20}}})
	def mutable_find6(col: PetCollection):
		return col.filter({}).find({"a": 5, "c": {"d.e": 20}})

	@test(expect={"a": 5, "b": 10, "c": {"d": {"e": 20}}})
	def mutable_find6(col: PetCollection):
		return col.filter({}).find({"a": 5, "c.d": {"e": 20}})

	@test(expect={"a": 5, "b": 10, "c": {"d": {"e": 20}}})
	def mutable_find6(col: PetCollection):
		return col.filter({}).find({"a": 5, "c": {"d": {"e": 20}}})

if __name__ == '__main__':
	print("Running tests...\n")
	Tests.init()
	passed, failed = Tests.run()
	if len(failed) == 0:
		print(f"{GREEN}All tests passed ({passed}){COLOREND}")
	else:
		print("\n".join(failed) + "\n")
		print(f"{RED}Tests failed ({len(failed)}/{len(failed) + passed}){COLOREND}")
